import numpy as np

from .cadence_features import (cad_prob, delta_t_hist, double_to_single_step,
                               normalize_hist, find_sorted_peaks, peak_bin,
                               peak_ratio)

from .common_functions import (maximum, median, max_slope,
                               median_absolute_deviation, minimum,
                               percent_beyond_1_std, percent_close_to_median,
                               skew, std, weighted_average)
from .amplitude import (amplitude, percent_amplitude, flux_percentile_ratio,
                        percent_difference_flux_percentile)
from .qso_model import (qso_fit, get_qso_log_chi2_qsonu,
                        get_qso_log_chi2nuNULL_chi2nu)
from .stetson import (stetson_j, stetson_k)

from .lomb_scargle import (lomb_scargle_model, get_lomb_frequency,
                           get_lomb_amplitude, get_lomb_rel_phase,
                           get_lomb_amplitude_ratio, get_lomb_frequency_ratio,
                           get_lomb_signif_ratio, get_lomb_lambda,
                           get_lomb_signif, get_lomb_varrat, get_lomb_trend,
                           get_lomb_y_offset)
from .lomb_scargle_fast import lomb_scargle_fast_period
from .num_alias import num_alias
from .periodic_model import (periodic_model, get_max_delta_mags,
                             get_min_delta_mags, get_model_phi1_phi2)
from .period_folding import (period_folding, get_fold2P_slope_percentile,
                             get_medperc90_2p_p, p2p_model,
                             get_p2p_scatter_2praw, get_p2p_scatter_over_mad,
                             get_p2p_scatter_pfold_over_mad,
                             get_p2p_ssqr_diff_over_var)
from .scatter_res_raw import scatter_res_raw


__all__ = ['CADENCE_FEATS', 'GENERAL_FEATS', 'LOMB_SCARGLE_FEATS',
           'generate_dask_graph', 'feature_categories', 'dask_feature_graph']

feature_categories = {
    'Cadence/Error': [
        'n_epochs','avg_err','med_err','std_err',
        'total_time','avgt','cads_std','mean',
        'cads_avg','cads_med','cad_probs_1',
        'cad_probs_10','cad_probs_20','cad_probs_30',
        'cad_probs_40','cad_probs_50','cad_probs_100',
        'cad_probs_500','cad_probs_1000','cad_probs_5000',
        'cad_probs_10000','cad_probs_50000','cad_probs_100000',
        'cad_probs_500000','cad_probs_1000000','cad_probs_5000000',
        'cad_probs_10000000','med_double_to_single_step',
        'avg_double_to_single_step','std_double_to_single_step',
        'all_times_nhist_numpeaks','all_times_nhist_peak_val',
        'all_times_nhist_peak_1_to_2','all_times_nhist_peak_1_to_3',
        'all_times_nhist_peak_2_to_3','all_times_nhist_peak_1_to_4',
        'all_times_nhist_peak_2_to_4','all_times_nhist_peak_3_to_4',
        'all_times_nhist_peak1_bin','all_times_nhist_peak2_bin',
        'all_times_nhist_peak3_bin','all_times_nhist_peak4_bin'
        ],

    'General': [
        'amplitude', 'flux_percentile_ratio_mid20',
        'flux_percentile_ratio_mid35', 'flux_percentile_ratio_mid50',
        'flux_percentile_ratio_mid65', 'flux_percentile_ratio_mid80',
        'max_slope', 'maximum', 'median', 'median_absolute_deviation',
        'minimum', 'percent_amplitude', 'percent_beyond_1_std',
        'percent_close_to_median',
        'percent_difference_flux_percentile', 'period_fast',
        'qso_log_chi2_qsonu', 'qso_log_chi2nuNULL_chi2nu', 'skew',
        'std', 'stetson_j', 'stetson_k', 'weighted_average'
        ],

    'Lomb-Scargle (Periodic)': [
        'fold2P_slope_10percentile', 'fold2P_slope_90percentile',
        'freq1_amplitude1', 'freq1_amplitude2', 'freq1_amplitude3',
        'freq1_amplitude4', 'freq1_freq', 'freq1_lambda',
        'freq1_rel_phase2', 'freq1_rel_phase3', 'freq1_rel_phase4',
        'freq1_signif', 'freq2_amplitude1', 'freq2_amplitude2',
        'freq2_amplitude3', 'freq2_amplitude4', 'freq2_freq',
        'freq2_rel_phase2', 'freq2_rel_phase3', 'freq2_rel_phase4',
        'freq3_amplitude1', 'freq3_amplitude2', 'freq3_amplitude3',
        'freq3_amplitude4', 'freq3_freq', 'freq3_rel_phase2',
        'freq3_rel_phase3', 'freq3_rel_phase4',
        'freq_amplitude_ratio_21', 'freq_amplitude_ratio_31',
        'freq_frequency_ratio_21', 'freq_frequency_ratio_31',
        'freq_model_max_delta_mags', 'freq_model_min_delta_mags',
        'freq_model_phi1_phi2', 'freq_n_alias',
        'freq_signif_ratio_21', 'freq_signif_ratio_31', 'freq_varrat',
        'freq_y_offset', 'linear_trend', 'medperc90_2p_p',
        'p2p_scatter_2praw', 'p2p_scatter_over_mad',
        'p2p_scatter_pfold_over_mad', 'p2p_ssqr_diff_over_var',
        'scatter_res_raw'
    ]
}

CADENCE_FEATS = feature_categories['Cadence/Error']
GENERAL_FEATS = feature_categories['General']
LOMB_SCARGLE_FEATS = feature_categories['Lomb-Scargle (Periodic)']


# See http://dask.pydata.org/en/latest/custom-graphs.html

dask_feature_graph = {
    'n_epochs': (len, 't'),
    'avg_err': (np.mean, 'e'),
    'med_err': (np.median, 'e'),
    'std_err': (np.std, 'e'),
    'total_time': (lambda x: np.max(x) - np.min(x), 't'),
    'avgt': (np.mean, 't'),
    'cads': (np.diff, 't'),
    'cads_std': (np.std, 'cads'),
    'mean': (np.mean, 'm'),
    'cads_avg': (np.mean, 'cads'),
    'cads_med': (np.median, 'cads'),
    'cad_probs_1': (cad_prob, 'cads', 1),
    'cad_probs_10': (cad_prob, 'cads', 10),
    'cad_probs_20': (cad_prob, 'cads', 20),
    'cad_probs_30': (cad_prob, 'cads', 30),
    'cad_probs_40': (cad_prob, 'cads', 40),
    'cad_probs_50': (cad_prob, 'cads', 50),
    'cad_probs_100': (cad_prob, 'cads', 100),
    'cad_probs_500': (cad_prob, 'cads', 500),
    'cad_probs_1000': (cad_prob, 'cads', 1000),
    'cad_probs_5000': (cad_prob, 'cads', 5000),
    'cad_probs_10000': (cad_prob, 'cads', 10000),
    'cad_probs_50000': (cad_prob, 'cads', 50000),
    'cad_probs_100000': (cad_prob, 'cads', 100000),
    'cad_probs_500000': (cad_prob, 'cads', 500000),
    'cad_probs_1000000': (cad_prob, 'cads', 1000000),
    'cad_probs_5000000': (cad_prob, 'cads', 5000000),
    'cad_probs_10000000': (cad_prob, 'cads', 10000000),
    'double_to_single_step': (double_to_single_step, 'cads'),
    'avg_double_to_single_step': (np.mean, 'double_to_single_step'),
    'med_double_to_single_step': (np.median, 'double_to_single_step'),
    'std_double_to_single_step': (np.std, 'double_to_single_step'),
    'delta_t_hist': (delta_t_hist, 't'),
    'delta_t_nhist': (normalize_hist, 'delta_t_hist', 'total_time'),
    'nhist_peaks': (find_sorted_peaks, 'delta_t_nhist'),
    'all_times_nhist_numpeaks': (len, 'nhist_peaks'),
    'all_times_nhist_peak_val': (np.max, 'delta_t_nhist'),
    'all_times_nhist_peak_1_to_2': (peak_ratio, 'nhist_peaks', 1, 2),
    'all_times_nhist_peak_1_to_3': (peak_ratio, 'nhist_peaks', 1, 3),
    'all_times_nhist_peak_2_to_3': (peak_ratio, 'nhist_peaks', 2, 3),
    'all_times_nhist_peak_1_to_4': (peak_ratio, 'nhist_peaks', 1, 4),
    'all_times_nhist_peak_2_to_4': (peak_ratio, 'nhist_peaks', 2, 4),
    'all_times_nhist_peak_3_to_4': (peak_ratio, 'nhist_peaks', 3, 4),
    'all_times_nhist_peak1_bin': (peak_bin, 'nhist_peaks', 1),
    'all_times_nhist_peak2_bin': (peak_bin, 'nhist_peaks', 2),
    'all_times_nhist_peak3_bin': (peak_bin, 'nhist_peaks', 3),
    'all_times_nhist_peak4_bin': (peak_bin, 'nhist_peaks', 4),

    # Standalone features (disconnected nodes)
    'amplitude': (amplitude, 'm'),
    'flux_percentile_ratio_mid20': (flux_percentile_ratio, 'm', 20),
    'flux_percentile_ratio_mid35': (flux_percentile_ratio, 'm', 35),
    'flux_percentile_ratio_mid50': (flux_percentile_ratio, 'm', 50),
    'flux_percentile_ratio_mid65': (flux_percentile_ratio, 'm', 65),
    'flux_percentile_ratio_mid80': (flux_percentile_ratio, 'm', 80),
    'maximum': (maximum, 'm'),
    'max_slope': (max_slope, 't', 'm'),
    'median': (median, 'm'),
    'median_absolute_deviation': (median_absolute_deviation, 'm'),
    'minimum': (minimum, 'm'),
    'percent_amplitude': (percent_amplitude, 'm'),
    'percent_beyond_1_std': (percent_beyond_1_std, 'm', 'e'),
    'percent_close_to_median': (percent_close_to_median, 'm'),
    'percent_difference_flux_percentile': (
        percent_difference_flux_percentile, 'm'),
    'skew': (skew, 'm'),
    'std': (std, 'm'),
    'stetson_j': (stetson_j, 'm'),
    'stetson_k': (stetson_k, 'm'),
    'weighted_average': (weighted_average, 'm', 'e'),

    # QSO model features
    'qso_model': (qso_fit, 't', 'm', 'e'),
    'qso_log_chi2_qsonu': (get_qso_log_chi2_qsonu, 'qso_model'),
    'qso_log_chi2nuNULL_chi2nu': (get_qso_log_chi2nuNULL_chi2nu,
                                  'qso_model'),

    # Fast Lomb-Scargle from Gatspy
    'period_fast': (lomb_scargle_fast_period, 't', 'm', 'e'),

    '_lomb_model': (lomb_scargle_model, 't', 'm', 'e'),
    # These could easily be programmatically generated, but this is more readable
    'freq1_freq': (get_lomb_frequency, '_lomb_model', 1),
    'freq2_freq': (get_lomb_frequency, '_lomb_model', 2),
    'freq3_freq': (get_lomb_frequency, '_lomb_model', 3),
    'freq1_amplitude1': (get_lomb_amplitude, '_lomb_model', 1, 1),
    'freq1_amplitude2': (get_lomb_amplitude, '_lomb_model', 1, 2),
    'freq1_amplitude3': (get_lomb_amplitude, '_lomb_model', 1, 3),
    'freq1_amplitude4': (get_lomb_amplitude, '_lomb_model', 1, 4),
    'freq2_amplitude1': (get_lomb_amplitude, '_lomb_model', 2, 1),
    'freq2_amplitude2': (get_lomb_amplitude, '_lomb_model', 2, 2),
    'freq2_amplitude3': (get_lomb_amplitude, '_lomb_model', 2, 3),
    'freq2_amplitude4': (get_lomb_amplitude, '_lomb_model', 2, 4),
    'freq3_amplitude1': (get_lomb_amplitude, '_lomb_model', 3, 1),
    'freq3_amplitude2': (get_lomb_amplitude, '_lomb_model', 3, 2),
    'freq3_amplitude3': (get_lomb_amplitude, '_lomb_model', 3, 3),
    'freq3_amplitude4': (get_lomb_amplitude, '_lomb_model', 3, 4),
    #        'freq1_rel_phase1': (get_lomb_rel_phase, 'lomb_model', 1, 1),
    'freq1_rel_phase2': (get_lomb_rel_phase, '_lomb_model', 1, 2),
    'freq1_rel_phase3': (get_lomb_rel_phase, '_lomb_model', 1, 3),
    'freq1_rel_phase4': (get_lomb_rel_phase, '_lomb_model', 1, 4),
    #        'freq2_rel_phase1': (get_lomb_rel_phase, 'lomb_model', 2, 1),
    'freq2_rel_phase2': (get_lomb_rel_phase, '_lomb_model', 2, 2),
    'freq2_rel_phase3': (get_lomb_rel_phase, '_lomb_model', 2, 3),
    'freq2_rel_phase4': (get_lomb_rel_phase, '_lomb_model', 2, 4),
    #        'freq3_rel_phase1': (get_lomb_rel_phase, 'lomb_model', 3, 1),
    'freq3_rel_phase2': (get_lomb_rel_phase, '_lomb_model', 3, 2),
    'freq3_rel_phase3': (get_lomb_rel_phase, '_lomb_model', 3, 3),
    'freq3_rel_phase4': (get_lomb_rel_phase, '_lomb_model', 3, 4),
    'freq_amplitude_ratio_21': (get_lomb_amplitude_ratio, '_lomb_model', 2),
    'freq_amplitude_ratio_31': (get_lomb_amplitude_ratio, '_lomb_model', 3),
    'freq_frequency_ratio_21': (get_lomb_frequency_ratio, '_lomb_model', 2),
    'freq_frequency_ratio_31': (get_lomb_frequency_ratio, '_lomb_model', 3),
    'freq_signif_ratio_21': (get_lomb_signif_ratio, '_lomb_model', 2),
    'freq_signif_ratio_31': (get_lomb_signif_ratio, '_lomb_model', 3),
    'freq1_lambda': (get_lomb_lambda, '_lomb_model'),
    'freq1_signif': (get_lomb_signif, '_lomb_model'),
    'freq_varrat': (get_lomb_varrat, '_lomb_model'),
    'linear_trend': (get_lomb_trend, '_lomb_model'),
    'freq_y_offset': (get_lomb_y_offset, '_lomb_model'),

     # Other features that operate on Lomb-Scargle residuals
    'freq_n_alias': (num_alias, '_lomb_model'),
    'scatter_res_raw': (scatter_res_raw, 't', 'm', 'e', '_lomb_model'),

    '_periodic_model': (periodic_model, '_lomb_model'),
    '_period_folded_model': (period_folding, 't', 'm', 'e', '_lomb_model'),

    'freq_model_max_delta_mags': (get_max_delta_mags, '_periodic_model'),
    'freq_model_min_delta_mags': (get_min_delta_mags, '_periodic_model'),
    'freq_model_phi1_phi2': (get_model_phi1_phi2, '_periodic_model'),
    'fold2P_slope_10percentile': (get_fold2P_slope_percentile,
                                  '_period_folded_model', 10),
    'fold2P_slope_90percentile': (get_fold2P_slope_percentile,
                                  '_period_folded_model', 90),
    'medperc90_2p_p': (get_medperc90_2p_p, '_period_folded_model'),

    '_p2p_model': (p2p_model, 't', 'm', 'freq1_freq'),
    'p2p_scatter_2praw': (get_p2p_scatter_2praw, '_p2p_model'),
    'p2p_scatter_over_mad': (get_p2p_scatter_over_mad, '_p2p_model'),
    'p2p_scatter_pfold_over_mad': (get_p2p_scatter_pfold_over_mad,
                                   '_p2p_model'),
    'p2p_ssqr_diff_over_var': (get_p2p_ssqr_diff_over_var, '_p2p_model')
}

def generate_dask_graph(t, m, e):
    full_graph = {'t': t, 'm': m, 'e': e}
    full_graph.update(dask_feature_graph)
    return full_graph


extra_feature_docs = {
    'n_epochs': 'Total number of observed values.',
    'avg_err': 'Mean of the error estimates.',
    'med_err': 'Median of error estimates.',
    'std_err': 'Standard deviation of the error estimates.',
    'total_time': 'Absolute difference between max and min of time values.',
    'avgt': 'Mean of the time values.',
    'mean': 'Mean of observed values.',
    'cads': 'List of differences between successive time values (`np.diff(t)`).',
    'cads_std': 'Standard deviation of `cads` (discrete difference between times).',
    'cads_avg': 'Mean value of `cads` (discrete difference between times).',
    'cads_med': 'Median value of `cads` (discrete difference between times).',
    'avg_double_to_single_step':
    'Mean value of ratios (t[i+2] - t[i]) / (t[i+2] - t[i+1]).',
    'med_double_to_single_step':
    'Median value of ratios (t[i+2] - t[i]) / (t[i+2] - t[i+1]).',
    'std_double_to_single_step':
    'Standard deviation of ratios (t[i+2] - t[i]) / (t[i+2] - t[i+1]).',
    'all_times_nhist_numpeaks':
    'Number of peaks (local maxima) in histogram of all possible delta_t\'s.',
    'all_times_nhist_peak_val':
    'Peak value in histogram of all possible delta_t\'s.'
}


feature_tags = {
    'n_epochs': ['Astronomy', 'General'],
    'avg_err': ['Astronomy', 'Error', 'General'],
    'med_err': ['Astronomy', 'Error', 'General'],
    'std_err': ['Astronomy', 'Error', 'General'],
    'total_time': ['Astronomy', 'General'],
    'avgt': ['Astronomy', 'General'],
    'cads': ['Astronomy', 'General', 'Cadence'],
    'cads_std': ['Astronomy', 'General', 'Cadence'],
    'mean': ['Astronomy', 'General'],
    'cads_avg': ['Astronomy', 'General', 'Cadence'],
    'cads_med': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_1': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_10': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_20': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_30': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_40': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_50': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_100': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_500': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_1000': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_5000': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_10000': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_50000': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_100000': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_500000': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_1000000': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_5000000': ['Astronomy', 'General', 'Cadence'],
    'cad_probs_10000000': ['Astronomy', 'General', 'Cadence'],
    'double_to_single_step': ['Astronomy', 'General', 'Cadence'],
    'avg_double_to_single_step': ['Astronomy', 'General', 'Cadence'],
    'med_double_to_single_step': ['Astronomy', 'General', 'Cadence'],
    'std_double_to_single_step': ['Astronomy', 'General', 'Cadence'],
    'delta_t_hist': ['Astronomy', 'General', 'Cadence'],
    'delta_t_nhist': ['Astronomy', 'General', 'Cadence'],
    'nhist_peaks': ['Astronomy', 'General', 'Cadence'],
    'all_times_nhist_numpeaks': ['Astronomy', 'General', 'Cadence'],
    'all_times_nhist_peak_val': ['Astronomy', 'General', 'Cadence'],
    'all_times_nhist_peak_1_to_2': ['Astronomy', 'General', 'Cadence'],
    'all_times_nhist_peak_1_to_3': ['Astronomy', 'General', 'Cadence'],
    'all_times_nhist_peak_2_to_3': ['Astronomy', 'General', 'Cadence'],
    'all_times_nhist_peak_1_to_4': ['Astronomy', 'General', 'Cadence'],
    'all_times_nhist_peak_2_to_4': ['Astronomy', 'General', 'Cadence'],
    'all_times_nhist_peak_3_to_4': ['Astronomy', 'General', 'Cadence'],
    'all_times_nhist_peak1_bin': ['Astronomy', 'General', 'Cadence'],
    'all_times_nhist_peak2_bin': ['Astronomy', 'General', 'Cadence'],
    'all_times_nhist_peak3_bin': ['Astronomy', 'General', 'Cadence'],
    'all_times_nhist_peak4_bin': ['Astronomy', 'General', 'Cadence'],

    # Standalone features (disconnected nodes)
    'amplitude': ['Astronomy', 'General'],
    'flux_percentile_ratio_mid20': ['Astronomy'],
    'flux_percentile_ratio_mid35': ['Astronomy'],
    'flux_percentile_ratio_mid50': ['Astronomy'],
    'flux_percentile_ratio_mid65': ['Astronomy'],
    'flux_percentile_ratio_mid80': ['Astronomy'],
    'maximum': ['Astronomy', 'General'],
    'max_slope': ['Astronomy', 'General'],
    'median': ['Astronomy', 'General'],
    'median_absolute_deviation': ['Astronomy', 'General'],
    'minimum': ['Astronomy', 'General'],
    'percent_amplitude': ['Astronomy', 'General'],
    'percent_beyond_1_std': ['Astronomy', 'General'],
    'percent_close_to_median': ['Astronomy', 'General'],
    'percent_difference_flux_percentile': ['Astronomy', 'General'],
    'skew': ['Astronomy', 'General'],
    'std': ['Astronomy', 'General'],
    'stetson_j': ['Astronomy', 'General'],
    'stetson_k': ['Astronomy', 'General'],
    'weighted_average': ['Astronomy', 'General'],

    # QSO model features
    'qso_model': ['Astronomy'],
    'qso_log_chi2_qsonu': ['Astronomy'],
    'qso_log_chi2nuNULL_chi2nu': ['Astronomy'],

    # Fast Lomb-Scargle from Gatspy
    'period_fast': ['Astronomy', 'Periodic', 'Lomb-Scargle'],

    '_lomb_model': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    # These could easily be programmatically generated, but this is more readable
    'freq1_freq': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq2_freq': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq3_freq': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq1_amplitude1': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq1_amplitude2': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq1_amplitude3': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq1_amplitude4': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq2_amplitude1': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq2_amplitude2': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq2_amplitude3': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq2_amplitude4': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq3_amplitude1': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq3_amplitude2': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq3_amplitude3': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq3_amplitude4': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    #        'freq1_rel_phase1': (get_lomb_rel_phase, 'lomb_model', 1, 1),
    'freq1_rel_phase2': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq1_rel_phase3': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq1_rel_phase4': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    #        'freq2_rel_phase1': (get_lomb_rel_phase, 'lomb_model', 2, 1),
    'freq2_rel_phase2': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq2_rel_phase3': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq2_rel_phase4': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    #        'freq3_rel_phase1': (get_lomb_rel_phase, 'lomb_model', 3, 1),
    'freq3_rel_phase2': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq3_rel_phase3': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq3_rel_phase4': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq_amplitude_ratio_21': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq_amplitude_ratio_31': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq_frequency_ratio_21': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq_frequency_ratio_31': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq_signif_ratio_21': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq_signif_ratio_31': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq1_lambda': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq1_signif': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq_varrat': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'linear_trend': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq_y_offset': ['Astronomy', 'Periodic', 'Lomb-Scargle'],

     # Other features that operate on Lomb-Scargle residuals
    'freq_n_alias': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'scatter_res_raw': ['Astronomy', 'Periodic', 'Lomb-Scargle'],

    '_periodic_model': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    '_period_folded_model': ['Astronomy', 'Periodic', 'Lomb-Scargle'],

    'freq_model_max_delta_mags': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq_model_min_delta_mags': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'freq_model_phi1_phi2': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'fold2P_slope_10percentile': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'fold2P_slope_90percentile': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'medperc90_2p_p': ['Astronomy', 'Periodic', 'Lomb-Scargle'],

    '_p2p_model': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'p2p_scatter_2praw': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'p2p_scatter_over_mad': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'p2p_scatter_pfold_over_mad': ['Astronomy', 'Periodic', 'Lomb-Scargle'],
    'p2p_ssqr_diff_over_var': ['Astronomy', 'Periodic', 'Lomb-Scargle']
}



import shutil
import tempfile
import uuid
from contextlib import contextmanager
from itertools import cycle, islice
from os.path import join as pjoin
import numpy as np
import pandas as pd

from cesium.time_series import TimeSeries


def sample_values(size=51, channels=1):
    times = np.sort(np.random.random(size))
    values = np.array([np.random.normal(size=size) for i in range(channels)])
    errors = np.array([np.random.exponential(size=size)
                       for i in range(channels)])
    if channels == 1:
        values = values[0]
        errors = errors[0]
    return times, values, errors


@contextmanager
def sample_ts_files(size, labels=[None]):
    temp_dir = tempfile.mkdtemp()
    paths = []
    for label in islice(cycle(labels), size):
        t, m, e = sample_values()
        name = str(uuid.uuid4())
        path = pjoin(temp_dir, '{}.npz'.format(name))
        ts = TimeSeries(t, m, e, label=label, path=path, name=name)
        ts.save(path)
        paths.append(path)

    yield paths

    shutil.rmtree(temp_dir)


def sample_featureset(size, n_channels=1, features=['mean'], labels=None,
                      names=None, meta_features=[]):
    ts_names = np.arange(size).astype('str')
    columns = pd.MultiIndex.from_tuples([(f, i) for f in features
                                         for i in range(n_channels)],
                                        names=['feature', 'channel'])
    fset = pd.DataFrame(np.random.random((size, len(features) * n_channels)),
                        columns=columns)
    if labels:
        labels = np.array(list(islice(cycle(labels), size)))
    if names:
        fset.index = names
    for feat in meta_features:
        fset[feat] = np.random.random(size)

    return fset, labels



import numpy as np
import numpy.testing as npt

from cesium.features import lomb_scargle
from cesium.features.graphs import LOMB_SCARGLE_FEATS
from cesium.features.tests.util import (generate_features, irregular_random,
                                        regular_periodic, irregular_periodic)


# These values are chosen because they lie exactly on the grid of frequencies
# searched by the Lomb Scargle optimization procedure
WAVE_FREQS = np.array([5.3, 3.3, 2.1])


def test_lomb_scargle_regular_single_freq():
    """Test Lomb-Scargle model features on regularly-sampled periodic data with
    one frequency/multiple harmonics. Estimated parameters should be very
    accurate in this case.
    """
    frequencies = np.hstack((WAVE_FREQS[0], np.zeros(len(WAVE_FREQS)-1)))
    amplitudes = np.zeros((len(frequencies),4))
    amplitudes[0,:] = [8,4,2,1]
    phase = 0.1
    times, values, errors = regular_periodic(frequencies, amplitudes, phase)
    all_lomb = generate_features(times, values, errors, LOMB_SCARGLE_FEATS)

    # Only test the first (true) frequency; the rest correspond to noise
    npt.assert_allclose(all_lomb['freq1_freq'], frequencies[0])

    # Hard-coded value from previous solution
    npt.assert_allclose(0.001996007984, all_lomb['freq1_lambda'], rtol=1e-7)

    for (i,j), amplitude in np.ndenumerate(amplitudes):
        npt.assert_allclose(amplitude,
                all_lomb['freq{}_amplitude{}'.format(i+1,j+1)], rtol=1e-2,
                    atol=1e-2)

    # Only test the first (true) frequency; the rest correspond to noise
    for j in range(1, amplitudes.shape[1]):
        npt.assert_allclose(phase*j*(-1**j),
            all_lomb['freq1_rel_phase{}'.format(j+1)], rtol=1e-2, atol=1e-2)

    # Frequency ratio not relevant since there is only; only test amplitude/signif
    for i in [2,3]:
        npt.assert_allclose(0., all_lomb['freq_amplitude_ratio_{}1'.format(i)], atol=1e-3)

    npt.assert_array_less(10., all_lomb['freq1_signif'])

    # Only one frequency, so this should explain basically all the variance
    npt.assert_allclose(0., all_lomb['freq_varrat'], atol=5e-3)

    # Exactly periodic, so the same minima/maxima should reoccur
    npt.assert_allclose(0., all_lomb['freq_model_max_delta_mags'], atol=1e-6)
    npt.assert_allclose(0., all_lomb['freq_model_min_delta_mags'], atol=1e-6)

    # Linear trend should be zero since the signal is exactly sinusoidal
    npt.assert_allclose(0., all_lomb['linear_trend'], atol=1e-4)

    folded_times = times % 1./(frequencies[0]/2.)
    sort_indices = np.argsort(folded_times)
    folded_times = folded_times[sort_indices]
    folded_values = values[sort_indices]

    # Residuals from doubling period should be much higher
    npt.assert_array_less(10., all_lomb['medperc90_2p_p'])

    # Slopes should be the same for {un,}folded data; use unfolded for stability
    slopes = np.diff(values) / np.diff(times)
    npt.assert_allclose(np.percentile(slopes,10),
        all_lomb['fold2P_slope_10percentile'], rtol=1e-2)
    npt.assert_allclose(np.percentile(slopes,90),
        all_lomb['fold2P_slope_90percentile'], rtol=1e-2)


def test_lomb_scargle_irregular_single_freq():
    """Test Lomb-Scargle model features on irregularly-sampled periodic data
    with one frequency/multiple harmonics. More difficult than
    regularly-sampled case, so we allow parameter estimates to be slightly
    noisy.
    """
    frequencies = np.hstack((WAVE_FREQS[0], np.zeros(len(WAVE_FREQS)-1)))
    amplitudes = np.zeros((len(WAVE_FREQS),4))
    amplitudes[0,:] = [8,4,2,1]
    phase = 0.1
    times, values, errors = irregular_periodic(frequencies, amplitudes, phase)
    all_lomb = generate_features(times, values, errors, LOMB_SCARGLE_FEATS)

    # Only test the first (true) frequency; the rest correspond to noise
    npt.assert_allclose(all_lomb['freq1_freq'], frequencies[0], rtol=1e-2)

    # Only test first frequency here; noise gives non-zero amplitudes for residuals
    for j in range(amplitudes.shape[1]):
        npt.assert_allclose(amplitudes[0,j],
                all_lomb['freq1_amplitude{}'.format(j+1)], rtol=5e-2, atol=5e-2)
        if j >= 1:
            npt.assert_allclose(phase*j*(-1**j),
                all_lomb['freq1_rel_phase{}'.format(j+1)], rtol=1e-1, atol=1e-1)

    npt.assert_array_less(10., all_lomb['freq1_signif'])

    # Only one frequency, so this should explain basically all the variance
    npt.assert_allclose(0., all_lomb['freq_varrat'], atol=5e-3)

    npt.assert_allclose(-np.mean(values), all_lomb['freq_y_offset'], rtol=5e-2)


def test_lomb_scargle_period_folding():
    """Tests for features derived from fitting a Lomb-Scargle periodic model
    and period-folding the data by the estimated period.
    """
    frequencies = np.hstack((WAVE_FREQS[0], np.zeros(len(WAVE_FREQS)-1)))
    amplitudes = np.zeros((len(WAVE_FREQS),4))
    amplitudes[0,:] = [8,4,2,1]
    phase = 0.1
    times, values, errors = irregular_periodic(frequencies, amplitudes, phase)
    all_lomb = generate_features(times, values, errors, LOMB_SCARGLE_FEATS)

    # Folding is numerically unstable so we need to use the exact fitted frequency
    freq_est = all_lomb['freq1_freq']
    # Fold by 1*period
    fold1ed_times = (times-times[0]) % (1./freq_est)
    sort_indices = np.argsort(fold1ed_times)
    fold1ed_times = fold1ed_times[sort_indices]
    fold1ed_values = values[sort_indices]
    # Fold by 2*period
    fold2ed_times = (times-times[0]) % (2./freq_est)
    sort_indices = np.argsort(fold2ed_times)
    fold2ed_times = fold2ed_times[sort_indices]
    fold2ed_values = values[sort_indices]

    npt.assert_allclose(np.sum(np.diff(fold2ed_values)**2) /
            np.sum(np.diff(values)**2), all_lomb['p2p_scatter_2praw'])
    npt.assert_allclose(np.sum(np.diff(values)**2) / ((len(values) - 1) *
        np.var(values)), all_lomb['p2p_ssqr_diff_over_var'])
    npt.assert_allclose(np.median(np.abs(np.diff(values))) /
            np.median(np.abs(values-np.median(values))),
            all_lomb['p2p_scatter_over_mad'])
    npt.assert_allclose(np.median(np.abs(np.diff(fold1ed_values))) /
                        np.median(np.abs(values-np.median(values))),
                        all_lomb['p2p_scatter_pfold_over_mad'])


def test_lomb_scargle_regular_multi_freq():
    """Test Lomb-Scargle model features on regularly-sampled periodic data with
    multiple frequencies, each with a single harmonic. Estimated parameters
    should be very accurate in this case.
    """
    frequencies = WAVE_FREQS
    amplitudes = np.zeros((len(frequencies),4))
    amplitudes[:,0] = [4,2,1]
    phase = 0.1
    times, values, errors = regular_periodic(frequencies, amplitudes, phase)
    all_lomb = generate_features(times, values, errors, LOMB_SCARGLE_FEATS)

    for i, frequency in enumerate(frequencies):
        npt.assert_allclose(frequency,
                all_lomb['freq{}_freq'.format(i+1)])

    for (i,j), amplitude in np.ndenumerate(amplitudes):
        npt.assert_allclose(amplitude,
                all_lomb['freq{}_amplitude{}'.format(i+1,j+1)],
                rtol=5e-2, atol=5e-2)

    for i in [2,3]:
        npt.assert_allclose(amplitudes[i-1,0] / amplitudes[0,0],
                all_lomb['freq_amplitude_ratio_{}1'.format(i)], atol=2e-2)

    npt.assert_array_less(10., all_lomb['freq1_signif'])


def test_lomb_scargle_irregular_multi_freq():
    """Test Lomb-Scargle model features on irregularly-sampled periodic data
    with multiple frequencies, each with a single harmonic. More difficult than
    regularly-sampled case, so we allow parameter estimates to be slightly
    noisy.
    """
    frequencies = WAVE_FREQS
    amplitudes = np.zeros((len(frequencies),4))
    amplitudes[:,0] = [4,2,1]
    phase = 0.1
    times, values, errors = irregular_periodic(frequencies, amplitudes, phase)
    all_lomb = generate_features(times, values, errors, LOMB_SCARGLE_FEATS)

    for i, frequency in enumerate(frequencies):
        npt.assert_allclose(frequency,
                all_lomb['freq{}_freq'.format(i+1)], rtol=1e-2)

    for (i,j), amplitude in np.ndenumerate(amplitudes):
        npt.assert_allclose(amplitude,
                all_lomb['freq{}_amplitude{}'.format(i+1,j+1)],
                rtol=1e-1, atol=1e-1)

    for i in [2,3]:
        npt.assert_allclose(amplitudes[i-1,0] / amplitudes[0,0],
                all_lomb['freq_amplitude_ratio_{}1'.format(i)], atol=2e-2)
        npt.assert_allclose(frequencies[i-1] / frequencies[0],
                all_lomb['freq_frequency_ratio_{}1'.format(i)], atol=5e-2)

    npt.assert_array_less(10., all_lomb['freq1_signif'])


def test_lomb_scargle_linear_trend():
    frequencies = np.hstack((WAVE_FREQS[0], np.zeros(len(WAVE_FREQS)-1)))
    amplitudes = np.zeros((len(WAVE_FREQS),4))
    amplitudes[0,:] = [8,4,2,1]
    phase = 0.1
    slope = 0.5

    # Estimated trend should be almost exact for noiseless data
    times, values, errors = regular_periodic(frequencies, amplitudes, phase)
    values += slope * times
    all_lomb = generate_features(times, values, errors, LOMB_SCARGLE_FEATS)
    npt.assert_allclose(slope, all_lomb['linear_trend'], rtol=1e-3)

    # Should still be close to true trend when noise is present
    times, values, errors = irregular_periodic(frequencies, amplitudes, phase)
    values += slope * times
    values += np.random.normal(scale=1e-3, size=len(times))
    all_lomb = generate_features(times, values, errors, LOMB_SCARGLE_FEATS)
    npt.assert_allclose(slope, all_lomb['linear_trend'], rtol=1e-1)


def test_scatter_res_raw():
    """Test feature that measures scatter of Lomb-Scargle residuals."""
    times, values, errors = irregular_random()
    lomb_model = lomb_scargle.lomb_scargle_model(times, values, errors)
    residuals = values - lomb_model['freq_fits'][0]['model']
    resid_mad = np.median(np.abs(residuals - np.median(residuals)))
    value_mad = np.median(np.abs(values - np.median(values)))
    f = generate_features(times, values, errors, ['scatter_res_raw'])
    npt.assert_allclose(f['scatter_res_raw'], resid_mad / value_mad, atol=3e-2)



import os
import shutil
import tarfile
from zipfile import ZipFile

import numpy as np
import numpy.testing as npt
import pandas as pd
from sklearn.datasets.base import Bunch

from cesium.datasets import andrzejak
from cesium.datasets import asas_training
import cesium.datasets.util as dsutil

try:
    from cStringIO import StringIO
except:
    from io import BytesIO as StringIO
try:
    import mock
except:
    import unittest.mock as mock


def urlpatch(*args):
    try:
        import urllib.request
        return mock.patch('urllib.request.urlopen', side_effect=args[0])
    except:
        import urllib2
        return mock.patch('urllib2.urlopen', side_effect=args[0])


class MockZipUrl(object):
    def __init__(self, url, num_columns=1):
        self.url = url
        self.num_columns = num_columns

    def read(self):
        out_buffer = StringIO()
        fname = os.path.basename(self.url)
        prefix = fname[0]
        if fname.endswith('.dat'):
            header = 'filename,class,meta1\n{}001,{},1.0\n'.format(prefix, prefix)
            out_buffer.write(header.encode('utf-8'))
        else:
            X = np.random.random((10, self.num_columns))
            text = '\n'.join(','.join(row) for row in X.astype('str'))
            if fname.endswith('.zip'):
                with ZipFile(out_buffer, 'w') as z:
                    z.writestr('{}001.txt'.format(prefix), text)
            elif fname.endswith('.tar.gz'):
                with tarfile.open(fileobj=out_buffer, mode='w:gz') as t:
                    tarinfo = tarfile.TarInfo('{}001.txt'.format(prefix))
                    tarinfo.size = len(text)
                    t.addfile(tarinfo, StringIO(text.encode('utf-8')))
        return out_buffer.getvalue()


def _mock_andrzejak_url(*args):
    return MockZipUrl(url=args[0], num_columns=1)


@urlpatch(_mock_andrzejak_url)
def test_fetch_andrzejak(self, tmpdir):
    """Test EEG data download."""
    num_files = len(andrzejak.ZIP_FILES)
    andrzejak.MD5SUMS = None
    data = andrzejak.fetch_andrzejak(str(tmpdir))
    assert(data['archive'].endswith('andrzejak.tar.gz')
           and os.path.exists(data['archive']))
    header = pd.read_csv(data['header'])
    npt.assert_array_equal(header['class'], data['classes'])
    assert(all(len(t) == len(m)
               for t, m in zip(data['times'], data['measurements'])))
    assert(len(data['times']) == num_files)
    assert(len(data['classes']) == num_files)
    assert(all(c in ['Z', 'N', 'O', 'S', 'F'] for c in data['classes']))


def _mock_asas_training_url(*args):
    return MockZipUrl(url=args[0], num_columns=3)


@urlpatch(_mock_asas_training_url)
def test_fetch_asas_training(self, tmpdir):
    """Test ASAS training data download."""
    num_files = 1
    asas_training.MD5SUMS = None
    data = asas_training.fetch_asas_training(str(tmpdir))
    assert(data['archive'].endswith('asas_training_set.tar.gz')
           and os.path.exists(data['archive']))
    header = pd.read_csv(data['header'])
    npt.assert_array_equal(header['class'], data['classes'])
    assert(all(len(t) == len(m) and len(m) == len(e)
               for t, m, e in zip(data['times'], data['measurements'],
                                  data['errors'])))
    assert(len(data['times']) == num_files)
    assert(len(data['classes']) == num_files)
    assert(len(data['metadata']) == num_files)
    assert(all(c in ['a'] for c in data['classes']))



import numpy as np


def amplitude(x):
    """Half the difference between the maximum and minimum magnitude."""
    return (np.max(x) - np.min(x)) / 2.0


# TODO old comment did not match code; is this the quantity we want to compute?
def percent_amplitude(x, base=10., exponent=-0.4):
    """Returns the largest distance from the median value, measured
    as a percentage of the median.

    Assumes data is log-scaled; by default we assume inputs are scaled as
    x=10^(-0.4*y), corresponding to units of magnitudes. Computations are
    performed on the corresponding linear-scale values.
    """
    linear_scale_data = base ** (exponent * x)
    y_max = np.max(linear_scale_data)
    y_min = np.min(linear_scale_data)
    y_med = np.median(linear_scale_data)
    return max(abs((y_max - y_med) / y_med), abs((y_med - y_min) / y_med))


def percent_difference_flux_percentile(x, base=10., exponent=-0.4):
    """Difference between the 95th and 5th percentiles of the data, expressed
    as a percentage of the median value.
    See Eyer (2005) arXiv:astro-ph/0511458v1, Evans & Belokurov (2005) (there
    the 98th and 2nd percentiles are used).

    Assumes data is log-scaled; by default we assume inputs are scaled as
    x=10^(-0.4*y), corresponding to units of magnitudes. Computations are
    performed on the corresponding linear-scale values.
    """
    linear_scale_data = base ** (exponent * x)
    y_95, y_50, y_5 = np.percentile(linear_scale_data, [95, 50, 5])
    return (y_95 - y_5) / y_50


def flux_percentile_ratio(x, percentile_range, base=10., exponent=-0.4):
    """A ratio of ((50+x) flux percentile - (50-x) flux percentile) /
    (95 flux percentile - 5 flux percentile), where x = percentile_range/2.

    Assumes data is log-scaled; by default we assume inputs are scaled as
    x=10^(-0.4*y), corresponding to units of magnitudes. Computations are
    performed on the corresponding linear-scale values.
    """
    linear_scale_data = base ** (exponent * x)
    y_high, y_low, y_95, y_5 = np.percentile(linear_scale_data,
            [50 + percentile_range / 2., 50 - percentile_range / 2., 95, 5])
    return (y_high - y_low) / (y_95 - y_5)



cdef extern from "_lomb_scargle.h":
     void lomb_scargle(int numt, int numf, int nharm, int detrend_order,
                       double psd[], double cn[], double wth[],
                       double sinx[], double cosx[], double sinx_step[],
                       double cosx_step[], double sinx_back[],
                       double cosx_back[], double sinx_smallstep[],
                       double cosx_smallstep[], double hat_matr[],
                       double hat_hat[], double hat0[],
                       double soln[], double chi0, double freq_zoom,
                       double psdmin, double tone_control,
                       double lambda0[], double lambda0_range[],
                       double Tr[], int ifreq[])



import os
import numpy as np
import pandas as pd
from . import util
from . import time_series
from .time_series import TimeSeries


__all__ = ['parse_ts_data', 'parse_headerfile', 'parse_and_store_ts_data']


# TODO more robust error handling
def parse_ts_data(filepath, sep=","):
    """Parses raw time series data file and returns an (n, 3) array of values.

    Data is expected as text in tabular format with separator `sep`. The output
    will always have three columns (time, measurement, error), even if the data
    file contains two or fewer:

    - For data containing three columns (time, measurement, error), all
      three are returned.
    - For data containing two columns, a dummy error column is added with
      value `time_series.DEFAULT_ERROR_VALUE`.
    - For data containing one column, a time column is also added with
      values evenly spaced from 0 to `time_series.DEFAULT_MAX_TIME`.

    Parameters
    ----------
    filename : str
        Path to raw time series data to be parsed.
    sep : str, optional
        Separator of columns in data file; defaults to ','.

    Returns
    -------
    np.ndarray
        3-column array of (time, measurement, error) values.
    """
    ts_data = np.loadtxt(filepath, delimiter=sep, ndmin=2)
    ts_data = ts_data[:, :3]  # Only using T, M, E
    if ts_data.shape[0] == 0 or ts_data.shape[1] == 0:
        raise ValueError("Incomplete or improperly formatted time series data"
                         " file provided.")
    elif ts_data.shape[1] == 1:
        ts_data = np.c_[np.linspace(0, time_series.DEFAULT_MAX_TIME,
                                    len(ts_data)),
                        ts_data,
                        np.repeat(time_series.DEFAULT_ERROR_VALUE,
                                  len(ts_data))]
    elif ts_data.shape[1] == 2:
        ts_data = np.c_[ts_data,
                        np.repeat(time_series.DEFAULT_ERROR_VALUE,
                                  len(ts_data))]
    return ts_data.T


def parse_headerfile(headerfile_path, files_to_include=None):
    """Parse header file containing classes/targets and meta-feature
    information.

    Parameters
    ----------
    headerfile_path : str
        Path to header file.

    files_to_include : list, optional
        If provided, only return the subset of rows from the header
        corresponding to the given filenames.

    Returns
    -------
    pandas.Series
        Class labels/targets from header file (if missing, all values are None)

    pandas.DataFrame
        Feature data from other columns besides filename, label (can be empty)
    """
    try:
        header = pd.read_csv(headerfile_path, comment='#')
    except:
        raise ValueError("Improperly formatted header file.")
    if 'filename' in header:
        header.index = [util.shorten_fname(str(f)) for f in header['filename']]
        header.drop('filename', axis=1, inplace=True)
    if files_to_include:
        short_fnames_to_include = [util.shorten_fname(str(f))
                                   for f in files_to_include]
        try:
            header = header.loc[short_fnames_to_include]
        except:
            raise ValueError("Incomplete header file: make sure your "
                              "header contains an entry for each time "
                              "series file in the uploaded archive, and "
                              "that the file names match the first column "
                              "of the header.")
    header.rename(columns={c: 'label' for c in ['label', 'target', 'class',
                                                'class_label']}, inplace=True)
    labels = (header.label if 'label' in header
              else pd.Series([None] * len(header.index), index=header.index))
    feature_data = header.drop(['label', 'class'], axis=1, errors='ignore')
    return labels, feature_data


def parse_and_store_ts_data(data_path, output_dir, header_path=None,
                            cleanup_archive=True, cleanup_header=True, sep=','):
    """Parses raw time series data from a single file or archive and loads
    metadata from header file (if applicable). Data is stored as files within
    `output_dir`, and the list of these paths is returned.

    Parameters
    ----------
    data_path : str
        Path to an individual time series file or tarball of multiple time
        series files to be used for feature generation.
    output_dir : str
        Directory in which time series files will be saved.
    header_path : str, optional
        Path to header file containing file names, labels/targets, and
        meta_features.
    cleanup_archive : bool, optional
        Boolean specifying whether to delete the uploaded data file/archive
        (defaults to True).
    cleanup_header : bool, optional
        Boolean specifying whether to delete the uploaded header file (defaults
        to True).
    sep : str, optional
        Separator of columns in data file; defaults to ','.

    Returns
    -------
    List of paths to time series files
    """
    with util.extract_time_series(data_path, cleanup_archive=cleanup_archive,
                                  cleanup_files=True) as ts_paths:
        short_fnames = [util.shorten_fname(f) for f in ts_paths]
        if header_path:
            labels, meta_features = parse_headerfile(header_path, ts_paths)
        else:
            labels = pd.Series([None] * len(short_fnames), index=short_fnames)
            meta_features = pd.DataFrame(index=short_fnames)

        all_time_series = []
        for ts_path in ts_paths:
            fname = util.shorten_fname(ts_path)
            t, m, e = parse_ts_data(ts_path, sep)
            ts_label = labels.loc[fname]
            ts_meta_features = meta_features.loc[fname]
            ts_path = '{}.npz'.format(fname)
            ts_path = os.path.join(output_dir, ts_path)
            ts = TimeSeries(t, m, e, ts_label, ts_meta_features, fname,
                            ts_path)
            ts.save(ts_path)
            all_time_series.append(ts_path)

    if header_path and cleanup_header:
        util.remove_files([header_path])

    return all_time_series



try:
    from StringIO import StringIO
except:
    from io import StringIO
import os
from os.path import join as pjoin
import shutil
import numpy as np
from cesium import data_management
from cesium import util

import numpy.testing as npt
import pytest


DATA_PATH = pjoin(os.path.dirname(__file__), "data")


def test_parse_ts_data(tmpdir):
    """Test time series data file parsing."""
    to_str = lambda X: StringIO('\n'.join([','.join(row)
                                           for row in X.astype(str).tolist()]))
    values = np.random.random((10, 3))

    t, m, e = data_management.parse_ts_data(to_str(values))
    npt.assert_allclose(t, values[:, 0])
    npt.assert_allclose(m, values[:, 1])
    npt.assert_allclose(e, values[:, 2])

    t, m, e = data_management.parse_ts_data(to_str(values[:, :2]))
    npt.assert_allclose(t, values[:, 0])
    npt.assert_allclose(m, values[:, 1])
    npt.assert_allclose(e[1:], e[:-1])  # constant

    t, m, e = data_management.parse_ts_data(to_str(values[:, :1]))
    npt.assert_allclose(np.diff(t), 1. / (len(values) - 1))
    npt.assert_allclose(m, values[:, 0])
    npt.assert_allclose(e[1:], e[:-1])  # constant

    with pytest.raises(ValueError):
        data_management.parse_ts_data(to_str(values[:, []]))


def test_parse_headerfile():
    """Test header file parsing."""
    headerfile_path = pjoin(DATA_PATH,
                            "asas_training_subset_classes_with_metadata.dat")

    labels, metadata = data_management.parse_headerfile(headerfile_path)
    npt.assert_array_equal(metadata.keys(), ["meta1", "meta2", "meta3"])
    npt.assert_equal(labels.loc["217801"], "Mira")
    npt.assert_almost_equal(metadata.loc["224635"].meta1, 0.330610932539)

    with pytest.raises(ValueError):
        labels, metadata = data_management.parse_headerfile(
            StringIO('test\n1,2\n3,4,5'))

    labels, metadata = data_management.parse_headerfile(headerfile_path,
        files_to_include=["217801"])
    npt.assert_array_equal(metadata.keys(), ["meta1", "meta2", "meta3"])
    npt.assert_equal(labels.loc["217801"], "Mira")

    with pytest.raises(ValueError):
        labels, metadata = data_management.parse_headerfile(headerfile_path,
            files_to_include=["111111111"])

    labels, metadata = data_management.parse_headerfile(headerfile_path,
        files_to_include=["217801"])
    npt.assert_array_equal(metadata.keys(), ["meta1", "meta2", "meta3"])
    npt.assert_equal(labels.loc["217801"], "Mira")


def test_parsing_and_saving(tmpdir):
    data_file_path = pjoin(DATA_PATH, "215153_215176_218272_218934.tar.gz")
    header_path = pjoin(DATA_PATH, "215153_215176_218272_218934_metadata.dat")
    time_series = data_management.parse_and_store_ts_data(
        data_file_path, str(tmpdir), header_path, cleanup_archive=False,
        cleanup_header=False)
    for ts in time_series:
        assert isinstance(ts, str)
        assert os.path.exists(ts)

    time_series = data_management.parse_and_store_ts_data(
        data_file_path, str(tmpdir), None, cleanup_archive=False,
        cleanup_header=False)
    for ts in time_series:
        assert isinstance(ts, str)
        assert os.path.exists(ts)



#! /usr/bin/env python

descr = """Machine Learning Time Series Platform

https://github.com/cesium-ml/cesium

"""

DISTNAME            = 'cesium'
DESCRIPTION         = 'Machine Learning Time-Series Platform'
LONG_DESCRIPTION    = descr
MAINTAINER          = 'cesium Team'
MAINTAINER_EMAIL    = 'stefanv@berkeley.edu'
URL                 = 'http://cesium-ml.org'
LICENSE             = 'Modified BSD'
DOWNLOAD_URL        = 'https://github.com/cesium-ml/cesium'
PYTHON_VERSION      = (3, 4)

import os
import sys

import setuptools
from distutils.command.build_py import build_py


def configuration(parent_package='', top_path=None):
    if os.path.exists('MANIFEST'):
         os.remove('MANIFEST')

    from numpy.distutils.misc_util import Configuration
    config = Configuration(None, parent_package, top_path)

    config.set_options(
            ignore_setup_xxx_py=True,
            assume_default_configuration=True,
            delegate_options_to_subpackages=True,
            quiet=True)

    config.add_subpackage('cesium')

    return config


with open('cesium/version.py') as fid:
    for line in fid:
        if line.startswith('version'):
            VERSION = line.split('=')[-1].strip(' \'"\n')
            break

with open('requirements.txt') as fid:
    INSTALL_REQUIRES = [l.split('#')[0].strip() for l in fid.readlines() if l
                        and not l.startswith('git')]
    INSTALL_REQUIRES = [pkg.replace('-', '_') for pkg in INSTALL_REQUIRES]


# requirements for those browsing PyPI
REQUIRES = [r.replace('>=', ' (>= ') + ')' if '=' in r else r
            for r in INSTALL_REQUIRES]
REQUIRES = [r.replace('==', ' (== ') for r in REQUIRES]


if __name__ == "__main__":
    try:
        from numpy.distutils.core import setup
        extra = {'configuration': configuration}
        # do not risk updating numpy
        INSTALL_REQUIRES = [r for r in INSTALL_REQUIRES if 'numpy' not in r]
    except ImportError:
        if len(sys.argv) >= 2 and ('--help' in sys.argv[1:] or
                                   sys.argv[1] in ('--help-commands',
                                                   'egg_info',
                                                   '--version',
                                                   'clean')):
            # For these actions, NumPy is not required.
            #
            # They are required to succeed without Numpy for example when
            # pip is used to install cesium when Numpy is not yet
            # present in the system.
            try:
                from setuptools import setup
            except ImportError:
                from distutils.core import setup
        else:
            print('To install cesium from source, you will need numpy.\n' +
                  'Install numpy with pip:\n' +
                  '  pip install numpy\n'
                  'Or using conda:\n'
                  '  conda install numpy\n'
                  'or use your operating system package manager.')
            sys.exit(1)

    setup(
        name=DISTNAME,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        url=URL,
        license=LICENSE,
        download_url=DOWNLOAD_URL,
        version=VERSION,

        classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: BSD License',
            'Programming Language :: C',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 3',
            'Topic :: Scientific/Engineering',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: POSIX',
            'Operating System :: Unix',
            'Operating System :: MacOS',
        ],

        install_requires=INSTALL_REQUIRES,
        setup_requires=['cython>=0.25'],
        requires=REQUIRES,
        configuration=configuration,
        packages=setuptools.find_packages(exclude=['doc']),
        include_package_data=True,
        zip_safe=False,
        cmdclass={'build_py': build_py},
    )



import numpy as np
from scipy.stats import norm
from scipy.linalg import solveh_banded, cholesky_banded
from scipy.special import gammaln, betainc, gammaincc


# TODO duplicate
def lprob2sigma(lprob):
    """Translates a log_e(probability) to units of Gaussian sigmas."""
    if lprob>-36.:
      sigma = norm.ppf(1.-0.5*np.exp(1.*lprob))
    else:
      sigma = np.sqrt( np.log(2./np.pi) - 2.*np.log(8.2) - 2.*lprob )
    return float(sigma)


def chol_inverse_diag(t):
    """Computes inverse of matrix given its Cholesky upper Triangular decomposition t.
    matrix form: ab[u + i - j, j] == a[i,j] (here u=1)
    (quick version: only calculates diagonal and neighboring elements)
    """
    (uu,nrows) = t.shape
    B = np.zeros((uu,nrows),dtype='float64')
    B[1,nrows-1] = 1.0/t[1,nrows-1]**2
    B[0,nrows-1] = -t[0,nrows-1]*B[1,nrows-1]/t[1,nrows-2]
    for j in reversed(range(nrows-1)):
      tjj = t[1,j]
      B[1,j] = (1.0/tjj-t[0,j+1]*B[0,j+1])/tjj
      B[0,j] = -t[0,j]*B[1,j]/t[1,j-1]
    return B


def qso_engine(time,data,error,ltau=3.,lvar=-1.7,sys_err=0.,return_model=False):
    """Calculates the fit quality of a damped random walk to a qso lightcurve.
    The formalism is from Rybicki & Press (1994; arXiv:comp-gas/9405004)

    Data are modelled with a covariance function
        Lij = 0.5*var*tau*exp(-|time_i-time_j|/tau) .

    Input:
        time - measurement times, typically days
        data - measured magnitudes
        error - uncertainty in measured magnitudes

    Output (dictionary):

        chi2/nu - classical variability measure
        chi2_qso/nu - for goodness of fit given fixed parameters
        chi2_qso/nu_extra - for parameter fitting, add to chi2/nu
        chi^2/nu_NULL - expected chi2/nu for non-qso variable

        signif_qso - significance chi^2/nu<chi^2/nu_NULL (rule out false alarm)
        signif_not_qso - significance chi^2/nu>1 (rule out qso)
        signif_vary - significance that source is variable
        class - resulting source type (ambiguous, not_qso, qso)

        model - time series prediction for each datum given all others (iff return_model==True)
        dmodel - model uncertainty, including uncertainty in data

    Notes:
        T = L^(-1)
        Data variance is D
        Full covariance C^(-1) = (L+D)^(-1) = T [T+D^(-1)]^(-1) D^(-1)
        Code takes advantage of the tridiagonality of T and T+D^(-1).
    """

    out_dict = {}
    out_dict['chi2_qso/nu']=999; out_dict['chi2_qso/nu_extra']=0.;
    out_dict['signif_qso']=0.; out_dict['signif_not_qso']=0.;  out_dict['signif_vary']=0.
    out_dict['chi2_qso/nu_NULL']=0.; out_dict['chi2/nu']=0.; out_dict['nu']=0
    out_dict['model']=[]; out_dict['dmodel']=[];
    out_dict['class']='ambiguous'

    lvar0 = np.log10(0.5) + lvar + ltau

    ln = len(data)
    dt = abs(time[1:]-time[:-1])

    # first make sure all dt>0
    g = np.where(dt>0.)[0]; lg = len(g)
    # must have at least 2 data points
    if lg <= 0:
        return out_dict

    if return_model:
        model = 1.*data; dmodel = -1.*error

    if lg < ln:
      dt = dt[g]
      gg = np.zeros(lg+1,dtype='int64'); gg[1:] = g+1
      dat = data[gg]; wt = 1./(sys_err**2+error[gg]**2)
      ln = lg+1
    else:
      dat = 1.*data
      wt = 1./(sys_err**2+error**2)

    out_dict['nu'] = ln-1.
    varx = np.var(dat)
    dat0 = (dat * wt).sum() / wt.sum()
    out_dict['chi2/nu'] = ((dat - dat0)**2 * wt).sum() / out_dict['nu']

    # define tridiagonal matrix T = L^(-1)
    # sparse matrix form: ab[u + i - j, j] == a[i,j]   i<=j, (here u=1)
    T = np.zeros((2,ln),dtype='float64')
    arg = dt*np.exp(-np.log(10)*ltau); ri = np.exp(-arg); ei = 1./(1./ri-ri)
    T[0,1:] = -ei; T[1,:-1] = 1.+ri*ei; T[1,1:] += ri*ei; T[1,ln-1] += 1.
    T0 = np.median(T[1,:]); T /= T0

    # equation for chi2_qso is [ (dat-x0) T Tp^(-1) D^(-1) (dat-x0) ]  , where Tp=T+D^(-1) and D^(-1)=wt
    fac = np.exp(np.log(10)*lvar0)/T0
    Tp = 1.*T
    Tp[1,:] += wt*fac
    # solve Tp*z=y for z (y=wt*dat)
    # This works for scipy __version__>='0.9.0' on anathem (20120809)
    b1 = (wt*dat).reshape((1,ln))
    b2 = b1.T
    #(Tpc,z) = solveh_banded(Tp,b2)
    z = solveh_banded(Tp,b2)
    Tpc = cholesky_banded(Tp) # the solveh_banded() function used to return the cholesky matrix, now we get seperately
    z = z.T
    z = z[0,:]
    c1 = wt.reshape((1,ln))
    c2 = c1.T
    #(Tpc,z0) = solveh_banded(Tp,c2)
    z0 = solveh_banded(Tp,c2)
    #HAS NOT CHANGED#Tpc2 = cholesky_banded(Tp)
    z0 = z0.T
    z0 = z0[0,:]


    #finally, get u=T*z
    u = T[1,:]*z; u[1:] += T[0,1:]*z[:-1]; u[:-1] += T[0,1:]*z[1:]
    u0 = T[1,:]*z0; u0[1:] += T[0,1:]*z0[:-1]; u0[:-1] += T[0,1:]*z0[1:]

    # magnitude offset x0, error = 1./sqrt(u0sum)
    u0sum = u0.sum(); x0 = u.sum()/u0sum;

    # fit statistic
    out_dict['chi2_qso/nu'] = np.dot(dat-x0,u-u0*x0)/out_dict['nu']

    # -2*log(likelihood) = chi2_qso + ldet_C + log(u0sum)
    #   first term: use chi2_qso/nu for goodness of fit with fixed parameters;
    #   all terms: use chi2_qso/nu + chi2_qso/nu_extra for fitting with variable parameters
    # get log of determinant for use later
    Tc = cholesky_banded(T)
    ldet_Tp = 2*np.log(Tpc[1,:]).sum()
    ldet_T = 2*np.log(Tc[1,:]).sum()
    ldet_C = ldet_Tp-ldet_T-np.log(wt).sum()
    out_dict['chi2_qso/nu_extra'] = (ldet_C + np.log(u0sum))/out_dict['nu']

    # get trace of C^(-1) for significance calculation
    Tpm = chol_inverse_diag(Tpc)
    diagC = T[1,:]*wt*Tpm[1,:]
    diagC[:-1] += T[0,1:]*wt[0:-1]*Tpm[0,1:]
    diagC[1:] += T[0,1:]*wt[1:]*Tpm[0,1:]
    TrC = diagC.sum()

    # significance in sigma units (large means false alarm unlikely)
    # (expected value of chi2_qso under the NULL hypothesis is TrC*varx)
    out_dict['chi2_qso/nu_NULL'] = TrC*varx/out_dict['nu']
    a=ln/2.; x = (out_dict['chi2_qso/nu']+1.e-8)/(out_dict['chi2_qso/nu_NULL']+out_dict['chi2_qso/nu']+1.e-8)
    prob = betainc(a,a,x)
    if prob <= 0:
      lprob = a*np.log(x) - np.log(a) + gammaln(2*a) - 2*gammaln(a)
    else:
      lprob = np.log(prob)
    out_dict['signif_qso'] = lprob2sigma(lprob)

    a=ln/2.; x = 1./(1.+out_dict['chi2_qso/nu'])
    prob = betainc(a,a,x)
    if prob <= 0:
      lprob = a*np.log(x) - np.log(a) + gammaln(2*a) - 2*gammaln(a)
    else:
      lprob = np.log(prob)
    out_dict['signif_not_qso'] = lprob2sigma(lprob)

    x = out_dict['chi2/nu']*out_dict['nu']
    prob = gammaincc(0.5*out_dict['nu'],0.5*x)
    if prob <= 0:
      lprob = (0.5*out_dict['nu']-1)*np.log(x) - 0.5*x - 0.5*out_dict['nu']*np.log(2) - gammaln(0.5*out_dict['nu'])
    else:
      lprob = np.log( prob )
    out_dict['signif_vary'] = lprob2sigma(lprob)

    if out_dict['signif_vary'] > 3:
        if out_dict['signif_qso'] > 3:
            out_dict['class']='qso'
        elif out_dict['signif_not_qso'] > 3:
            out_dict['class']='not_qso'

    # best-fit model for the lightcurve
    if return_model:
      model[gg] = dat - (u-u0*x0)/diagC
      dmodel[gg] = 1./np.sqrt(diagC)
      out_dict['model'] = model
      out_dict['dmodel'] = dmodel

    return out_dict


def qso_fit(time, data, error, filter='g', mag0=19., sys_err=0.0, return_model=False):
    """Best-fit qso model determined for Sesar Strip82, ugriz-bands (default r).
    See additional notes for underlying code qso_engine.

    Input:
        time - measurement times [days]
        data - measured magnitudes in single filter (also specified)
        error - uncertainty in measured magnitudes

    Output:
        chi^2/nu - classical variability measure
        chi^2_qso/nu - fit statistic
        chi^2_qso/nu_NULL - expected fit statistic for non-qso variable

        signif_qso - significance chi^2/nu<chi^2/nu_NULL (rule out false alarm)
        signif_not_qso - significance chi^2/nu>1 (rule out qso)
        signif_vary - significance that source is variable at all
        class - source type (ambiguous, not_qso, qso)

        model - time series prediction for each datum given all others (iff return_model==True)
        dmodel - model uncertainty, including uncertainty in data

    Note on use (i.e., how class is defined):

          (0) signif_vary < 3: ambiguous, else
          (1) signif_qso > 3: qso, else
          (2) signif_not_qso > 3: not_qso
    """

    data = data.copy() - np.median(data) + mag0
    pars={}
    pars['u'] = [-3.90, 0.12, 2.73, -0.02]
    pars['g'] = [-4.10, 0.14, 2.92, -0.07]
    pars['r'] = [-4.34, 0.20, 3.12, -0.15]
    pars['i'] = [-4.23, 0.05, 2.83,  0.07]
    pars['z'] = [-4.44, 0.13, 3.06, -0.07]

    par = pars[filter.lower()]
    lvar = par[0]+par[1]*(mag0-19.)
    ltau = par[2]+par[3]*(mag0-19.)

    adict = qso_engine(time, data, error, ltau=ltau, lvar=lvar, return_model=return_model, sys_err=sys_err)

    out_dict={}
    out_dict['lvar']=lvar
    out_dict['ltau']=ltau
    out_dict['chi2/nu']=adict['chi2/nu']
    out_dict['nu'] = adict['nu']
    out_dict['chi2_qso/nu']=adict['chi2_qso/nu']
    out_dict['chi2_qso/nu_NULL']=adict['chi2_qso/nu_NULL']
    out_dict['signif_qso']=adict['signif_qso']
    out_dict['signif_not_qso']=adict['signif_not_qso']
    out_dict['signif_vary']=adict['signif_vary']
    out_dict['class']=adict['class'];
    out_dict['chi2qso_nu_nuNULL_ratio'] = out_dict['chi2_qso/nu'] / out_dict['chi2_qso/nu_NULL']

    ### Nat has converged upon the following being the most significant featues,
    #    Joey believes it is best to jut use these features only (so now the others are disabled in
    #       __init__.py and qso_extractor.py
    out_dict['log_chi2_qsonu'] = np.log(out_dict['chi2_qso/nu'])
    out_dict['log_chi2nuNULL_chi2nu'] = np.log(out_dict['chi2_qso/nu_NULL'] / out_dict['chi2_qso/nu'])
    ###

    if return_model:
      out_dict['model'] = adict['model']
      out_dict['dmodel'] = adict['dmodel']

    return out_dict


def get_qso_log_chi2_qsonu(qso_model):
    """Natural log of goodness of fit of qso-model given fixed parameters."""
    return qso_model['log_chi2_qsonu']


def get_qso_log_chi2nuNULL_chi2nu(qso_model):
    """Natural log of expected chi2/nu for non-qso variable."""
    return qso_model['log_chi2nuNULL_chi2nu']



import numpy as np


# TODO is this really a useful feature?
def num_alias(lomb_model):
    """ Here we check for "1-day" aliases in ASAS / Deboss sources."""
    alias = [{'per':1.,
              'p_low':0.92,
              'p_high':1.08,
              'alpha_1':8.191855,
              'alpha_2':-7.976243},
             {'per':0.5,
              'p_low':0.48,
              'p_high':0.52,
              'alpha_1':2.438913,
              'alpha_2':0.9837243},
             {'per':0.3333333333,
              'p_low':0.325,
              'p_high':0.342,
              'alpha_1':2.95749,
              'alpha_2':-4.285432},
             {'per':0.25,
              'p_low':0.245,
              'p_high':0.255,
              'alpha_1':1.347657,
              'alpha_2':2.326338}]

    count = 0
    for freq_dict in lomb_model['freq_fits']:
        period = 1. / freq_dict['freq']
        for a in alias:
            cutoff = (a['alpha_1'] / np.power(np.abs(period - a['per']), 0.25)
                      + a['alpha_2'])
            if (period >= a['p_low'] and period <= a['p_high'] and
                freq_dict['signif'] < cutoff):
                count += 1  # this frequency has a "1 day" alias (or 0.5 or 0.33
                break  # only need to do this once per period, if an alias is found.
    return count



import os
import numpy as np
from Cython.Build import cythonize

base_path = os.path.abspath(os.path.dirname(__file__))


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('features', parent_package, top_path)

    cythonize(os.path.join(base_path, '_lomb_scargle.pyx'))

    config.add_extension('_lomb_scargle', '_lomb_scargle.c',
                         include_dirs=[np.get_include()])

    return config

if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())
